//---------------------------------------------------------------------------

#ifndef AllometryH
#define AllometryH


#include <math.h>
#include "TreePopulation.h"

/**
 Allometry Class - Version 1.0
 This handles allometric calculations for trees.  None of these functions
 check for over- or underflow.

 Copyright 2003 Charles D. Canham.
 @author Lora E. Murphy

 <br>Edit history:
 <br>-----------------
 <br>April 28, 2004 - Submitted as beta (LEM)
*/
class clAllometry {

 public:
/**
  Constructor.
  @param p_oPop Tree population object.
*/
 clAllometry(clTreePopulation *p_oPop) {mp_oTreePop = p_oPop;};

/**
* Calculates height according to the adult allometry equation. This equation is:
* Height = 1.35 + (H<sub>1</sub> - 1.35)*(1 - e<sup>-B * DBH</sup>) where
* Height is height in meters, H<sub>1</sub> is max tree height, B is slope
* of asymptotic height, and DBH is tree DBH in cm.
* @param fDbh DBH in cm for which to calculate tree height.
* @param iSpecies Species of tree for which to calculate height.
* @return Height in meters.
* @throw ModelErr If the species isn't recognized.
*/
 float CalcAdultHeight( float fDbh, int iSpecies );

/**
* Calculates DBH according to the adult allometry equation.  This equation
* (which is just the height equation switched around) is:
* DBH = log(1-((Height-1.35)/(H<sub>1</sub>-1.35)))/-B where
* Height is height in meters, H<sub>1</sub> is max tree height, B is slope
* of asymptotic height, and DBH is tree DBH in cm.
* @param fHeight Height in meters.
* @param iSpecies Species of tree for which to calculate height.
* @return DBH in cm.
* @throw ModelErr If the species isn't recognized.
*/
 float CalcAdultDbh( float fHeight, int iSpecies );

/**
* Calculates canopy radius according to the adult allometry equation.  This
* equation is: Canrad = C<sub>1</sub> * DBH<sup>xi</sup> where
* Canrad is the canopy radius in meters, xi is the canopy radius
* to DBH ratio, DBH is the DBH of the tree in cm, and C<sub>exp</sub> is the
* canopy radius exponent.  The value of Canrad cannot exceed MAX_CANOPY_RADIUS.
* @param fDbh DBH in cm for which to calculate tree height.
* @param iSpecies Species of tree for which to calculate height.
* @return Canopy radius in meters, up to a maximum value of MAX_CANOPY_RADIUS.
* @throw ModelErr If the species isn't recognized.
*/
 float CalcAdultCanRadius( float fDbh, int iSpecies );

/**
* Calculates adult canopy depth according to the adult allometry equation. This
* equation is: CanDepth = C<sub>2</sub> * Height<sup>psi</sup> where
* CanDepth is the depth of the canopy in meters, C<sub>2</sub> is the fraction
* of tree height in canopy, Height is the tree's height in meters, and
* psi is the canopy depth exponent.
* @param fHeight Height in meters.
* @param iSpecies Species of tree for which to calculate height.
* @return Adult canopy depth in meters.
* @throw ModelErr If the species isn't recognized.
*/
 float CalcAdultCanDepth(float fHeight, int iSpecies);

/**
* Calculates height according to the sapling allometry equation (which is
* currently the same as the adult equation).  This equation is:
* Height = 1.35 + (H<sub>1</sub> - 1.35)*(1 - e<sup>-B * DBH</sup>) where
* Height is height in meters, H<sub>1</sub> is max tree height, B is slope
* of asymptotic height, and DBH is tree DBH in cm.
* @param fDbh DBH in cm for which to calculate tree height.
* @param iSpecies Species of tree for which to calculate height.
* @return Height in meters.
* @throw ModelErr If the species isn't recognized.
*/
 float CalcSaplingHeight(float fDbh, int iSpecies)
   {return CalcAdultHeight(fDbh, iSpecies);};

/**
* Calculates DBH according to the sapling allometry equation (which is currently
* the same as the adult equation).  This equation (which is just the height
* equation switched around) is:
* DBH = log(1-((Height-1.35)/(H<sub>1</sub>-1.35)))/-B where
* Height is height in meters, H<sub>1</sub> is max tree height, B is slope
* of asymptotic height, and DBH is tree DBH in cm.
* @param fHeight Height in meters.
* @param iSpecies Species of tree for which to calculate height.
* @return DBH in cm.
* @throw ModelErr If the species isn't recognized.
*/
 float CalcSaplingDbh( float fHeight, int iSpecies )
   {return CalcAdultDbh(fHeight, iSpecies);};

/**
* Calculates canopy radius according to sapling the allometry equation (which is
* currently the same as the adult equation).  This
* equation is: Canrad = C<sub>1</sub> * DBH<sup>xi</sup> where
* Canrad is the canopy radius in meters, xi is the canopy radius
* to DBH ratio, DBH is the DBH of the tree in cm, and C<sub>exp</sub> is the
* canopy radius exponent.  The value of Canrad cannot exceed MAX_CANOPY_RADIUS.
* @param fDbh DBH in cm for which to calculate tree height.
* @param iSpecies Species of tree for which to calculate height.
* @return Canopy radius in meters, up to a maximum value of MAX_CANOPY_RADIUS.
* @throw ModelErr If the species isn't recognized.
*/
 float CalcSaplingCanRadius( float fDbh, int iSpecies )
   {return CalcAdultCanRadius(fDbh, iSpecies);};

/**
* Calculates canopy depth according to the sapling allometry equation (which is
* currently the same as the adult equation).  This
* equation is: CanDepth = C<sub>2</sub> * Height<sup>psi</sup> where
* CanDepth is the depth of the canopy in meters, C<sub>2</sub> is the fraction
* of tree height in canopy, Height is the tree's height in meters, and
* psi is the canopy depth exponent.
* @param fHeight Height in meters.
* @param iSpecies Species of tree for which to calculate height.
* @return Adult canopy depth in meters.
* @throw ModelErr If the species isn't recognized.
*/
 float CalcSaplingCanDepth(float fHeight, int iSpecies)
   {return CalcAdultCanDepth(fHeight, iSpecies);};

/**
* Calculates height according to seedling allometry equation.  This equation is:
* height = 0.1 + 30*(1 - e<sup>-alpha * diam<sub>10</sub>)</sup>) where
* height is tree height in m, alpha is the slope of the height -
* diam<sub>10</sub> relationship, and diam<sub>10</sub> is the tree's diameter
* at 10 cm height.
* @param fDiam10 Diameter at 10 cm in cm for which to calculate tree height.
* @param iSpecies Species of tree for which to calculate height.
* @return Height in meters.
* @throw ModelErr If the species isn't recognized.
*/
 float CalcSeedlingHeight( float fDiam10, int iSpecies );

/**
* Calculates diameter at 10 cm according to seedling allometry equation (which
* is just the height equation switched around).  The equation is:
* diam<sub>10</sub> = (log(1-((height-0.1)/30)))/-alpha where
* height is tree height in m, alpha is the slope of the height -
* diam<sub>10</sub> relationship, and diam<sub>10</sub> is the tree's diameter
* at 10 cm height.
* @param fHeight Height in meters.
* @param iSpecies Species of tree for which to calculate height.
* @return Diameter at 10 cm in cm.
* @throw ModelErr If the species isn't recognized.
*/
 float CalcSeedlingDiam10( float fHeight, int iSpecies );

/**
* Converts a diameter at 10 cm value to a DBH.  The equation is:
* DBH = diam<sub>10</sub> * R where DBH is the DBH in cm, diam<sub>10</sub> is
* the diameter at 10 cm height in cm, and R is the DBH to diam<sub>10</sub>
* ratio.
* @param fDiam10 Diameter at 10 cm in cm.
* @param iSpecies Species of tree.
* @return DBH value in cm.
* @throw ModelErr If the species isn't recognized.
*/
 float ConvertDiam10ToDbh( float fDiam10, int iSpecies );

/**
* Converts a DBH value to a diameter at 10 cm value.  The equation is:
* diam<sub>10</sub> = DBH / R where DBH is the DBH in cm, diam<sub>10</sub> is
* the diameter at 10 cm height in cm, and R is the DBH to diam<sub>10</sub>
* ratio.
* @param fDbh DBH in cm.
* @param iSpecies Species of tree.
* @return Diameter at 10 cm value in cm.
* @throw ModelErr If the species isn't recognized.
*/
 float ConvertDbhToDiam10( float fDbh, int iSpecies );

 protected:
 clTreePopulation *mp_oTreePop; /**<Tree population object - this is so
   errors can be reported*/

}; //end of class clAllometry


//---------------------------------------------------------------------------
#endif
