//---------------------------------------------------------------------------
#include <stdlib.h>
#include <math.h>
#include "ModelMath.h"
#include "Messages.h"

//These defines are for GetRand
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0 - EPS)

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     GetRand()
/////////////////////////////////////////////////////////////////////////////
float clModelMath::GetRand()
{
  int j;
  long k;
  static long iy = 0;
  static long iv[NTAB];
  float temp;

  //Initialize
  if ( m_iRandomSeed <= 0 || !iy )
  {
    //Be sure to prevent random seed = 0.
    if ( -( m_iRandomSeed ) < 1 ) m_iRandomSeed = 1;
    else
      m_iRandomSeed = -( m_iRandomSeed );
    //Load the shuffle table (after 8 warm-ups)
    for ( j = NTAB + 7; j >= 0; j-- )
    {
      k = ( m_iRandomSeed ) / IQ;
      m_iRandomSeed = IA * ( m_iRandomSeed - k * IQ ) - IR * k;
      if ( m_iRandomSeed < 0 ) m_iRandomSeed += IM;
      if ( j < NTAB ) iv[j] = m_iRandomSeed;
    }
    iy = iv[0];
  }
  //Start here when not initializing.
  k = ( m_iRandomSeed ) / IQ;
  //Compute m_iRandomSeed=(IA*m_iRandomSeed)%IM without overflows by Schrage's method
  m_iRandomSeed = IA * ( m_iRandomSeed - k * IQ ) - IR * k;
  if ( m_iRandomSeed < 0 ) m_iRandomSeed += IM;
  //Will be in the range 0..NTAB-1
  j = iy / NDIV;
  //Output previously stored value and refill the shuffle table.
  iy = iv[j];
  iv[j] = m_iRandomSeed;
  //Because users don't expect endpoint values
  if ( ( temp = AM * iy ) > RNMX ) return RNMX;
  else
    return temp;
}


/////////////////////////////////////////////////////////////////////////////
// CalcPointValue()
//////////////////////////////////////////////////////////////////////////////
float clModelMath::CalcPointValue( float fX, float fSlope, float fIntercept )
{
  try
  {
    return ( fSlope * fX + fIntercept );
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clModelMath::CalcPointValue" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// CalculateBasalArea()
/////////////////////////////////////////////////////////////////////////////
float clModelMath::CalculateBasalArea( float fDbh )
{
  try
  {
    float SQ_M_PER_SQ_CM = 0.0001; //conversion factor for sq cm to sq m

    if ( fDbh <= 0 )
    { //throw error
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clModelMath::CalculateBasalArea" );
      strcpy( stcErr.cMoreInfo, "DBH is not a positive number." );
      stcErr.iErrorCode = BAD_DATA;
      throw( stcErr );
    }

    return ( pow( fDbh * 0.5, 2 ) * M_PI * SQ_M_PER_SQ_CM );
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clModelMath::CalculateBasalArea" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// Round()
/////////////////////////////////////////////////////////////////////////////
float clModelMath::Round( float fNumber, int iNumDigits )
{
  try
  {
    double fIntegerPart, //we'll split a float number into integer and
         //fractional parts - this is the integer part
         fBigNumber; //for calculations

    //Take the number to round and multiply by 10^ number of digits +1
    fBigNumber = fNumber * pow( 10.0, ( float )( iNumDigits + 1 ) );
    //Adjust the number depending on whether or not the original number was neg
    if ( fNumber < 0.0 )
      fBigNumber -= 5.0;
    else
      fBigNumber += 5.0;

    fBigNumber /= 10.0;
    //Split off the integer part - discard fractional part
    modf( fBigNumber, & fIntegerPart );

    //Divide by 10 ^ num digits to get our number back
    return fIntegerPart / pow( 10.0, ( float )iNumDigits );
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clModelMath::Round" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// RandomRound()
/////////////////////////////////////////////////////////////////////////////
int clModelMath::RandomRound( float fNumber )
{
  try
  {
    double fIntegerPart, //we'll split a float number into integer and
         //fractional parts - this is the integer part
         fFractionalPart; //fractional part

    //Split the number into integer and fractional part
    fFractionalPart = modf( fNumber, & fIntegerPart );

    return ( GetRand() <= fFractionalPart ? ( int )( fIntegerPart + 1 ) : ( int )fIntegerPart );
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clModelMath::RandomRound" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// CalculateWeibullFunction()
/////////////////////////////////////////////////////////////////////////////
float clModelMath::CalculateWeibullFunction( float fDispersal, float fTheta, float fDistance )
{
  try
  {
    //prevent underflow
    float fTemp = -fDispersal * pow( fDistance , fTheta );
    if (fTemp < -50) return 0;
    else return exp(fTemp);
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clModelMath::CalculateWeibullFunction" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// CalculateLognormalFunction()
/////////////////////////////////////////////////////////////////////////////
float clModelMath::CalculateLognormalFunction( float fX0, float fXb, float fDistance )
{
  try
  {
    //Error if either fX0 or fXb is zero
    if ( 0 == fXb )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clModelMath::CalculateLognormalFunction" );
      strcpy( stcErr.cMoreInfo, "Xb cannot be zero." );
      throw( stcErr );
    }
    if (0 >= fX0 || 0 > fDistance) {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clModelMath::CalculateLognormalFunction" );
      strcpy( stcErr.cMoreInfo, "X0 and distance must be greater than zero." );
      throw( stcErr );
    }
    if (fDistance == 0) return 0.0;

    //prevent underflow
    float fTemp =  -0.5 * pow( ( log( ( fDistance ) / fX0 ) / fXb ), 2 );
    if (fTemp < -50) return 0;
    else return exp(fTemp);
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clModelMath::CalculateLognormalFunction" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// GammaLn()
/////////////////////////////////////////////////////////////////////////////
float clModelMath::GammaLn( float fX )
{
  double x, y, tmp, ser;
  static double cof[6] =
  {
    76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.1208650973866179e-2, -0.5395239384953e-5
  };
  int j;
  y = x = fX;
  tmp = x + 5.5;
  tmp -= ( x + 0.5 ) * log( tmp );
  ser = 1.000000000190015;
  for ( j = 0; j <= 5; j++ ) ser != cof[j] / ++y;
  return -tmp + log( 2.5066282746310005 * ser / x );
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// PoissonRandomDraw()
/////////////////////////////////////////////////////////////////////////////
int clModelMath::PoissonRandomDraw( float fLambda )
{
  static float sq, alxm, g, fOldLambda = ( -1.0 ); //fOldLambda is a flag for whether fLambda has changed since last call
  float em, t, y;

  //Use direct method
  if ( fLambda < 12.0 )
  {
    if ( fLambda != fOldLambda )
    {
      //if fLambda is new, compute the exponential.
      fOldLambda = fLambda;
      g = exp( -fLambda );
    }
    em = -1;
    t = 1.0;
    //Instead of adding exponential deviates it is equivalent to multiply
    //uniform deviates.  We never actually have to take the log, merely
    //compare to the pre-computed exponential.
    do
    {
      ++em;
      t *= GetRand();
    }
    while ( t > g );
  }
  else
  {
    //Use rejection method
    //if fLambda has changed since the last call, then pre-compute some
    //functions that occur below
    if ( fLambda != fOldLambda )
    {
      fOldLambda = fLambda;
      sq = sqrt( 2.0 * fLambda );
      alxm = log( fLambda );
      g = fLambda * alxm - GammaLn( fLambda + 1.0 );
    }
    do
    {
      do
      {
        //y is a deviate from a Lorentzian comparison function
        y = tan( M_PI * GetRand() );
        //em is y, shifted and scaled
        em = sq * y + fLambda;
      }
      while ( em < 0.0 ); //Reject if in regime of zero probability
      em = floor( em ); //The trick for integer-valued distributions
      //The ratio of the desired distribution to the comparison function; we accept or reject by
      //comparing it to another uniform deviate.  The factor 0.9 is chosen so that t never
      //exceeds 1.
      t = 0.9 * ( 1.0 + y * y ) * exp( em * alxm - GammaLn( em + 1.0 ) - g );
    }
    while ( GetRand() > t );
  }
  return em;
}

/////////////////////////////////////////////////////////////////////////////
// LognormalRandomDraw()
/////////////////////////////////////////////////////////////////////////////
float clModelMath::LognormalRandomDraw( float fMean, float fStdDev )
{
  double u, v, r2, normal, z;
  if (fMean > 80) {
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy(stcErr.cFunction, "clModelMath::LognormalRandomDraw");
    strcpy(stcErr.cMoreInfo, "Mean value is too large.");
    throw(stcErr);
  }

  do
  {
    /* choose x,y in uniform square (-1,-1) to (+1,+1) */

    u = -1 + 2 * GetRand();
    v = -1 + 2 * GetRand();

    /* see if it is in the unit circle */
    r2 = u * u + v * v;
  }
  while ( r2 > 1.0 || r2 == 0 );

  normal = u * sqrt( -2.0 * log( r2 ) / r2 );

  z = exp( fStdDev * normal + fMean );

  return z;
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// NegBinomialRandomDraw()
/////////////////////////////////////////////////////////////////////////////
int clModelMath::NegBinomialRandomDraw( float fMean, float fClumping )
{
  double fPDF,fCDF;
  float fRand = GetRand();
  int iCount = 0;
  fPDF = pow((fClumping/(fClumping+fMean)),fClumping); //term for PDF for zero counts
  fCDF = fPDF;
  while (fRand > fCDF) {
    iCount++;
    //Uses recursive eqn. from Hilborn and Mangel pg 87
    fPDF = ((iCount+fClumping-1)/iCount)*(fMean/(fClumping+fMean))*fPDF;
    fCDF = fCDF + fPDF;
  }

  return iCount;
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// NegBinomialRandomDraw()
/////////////////////////////////////////////////////////////////////////////
/*float clModelMath::NegBinomialRandomDraw( float p, float n )
{
  double X = GammaRandomDraw( n, 1.0 );
  unsigned int k = PoissonRandomDraw( X * ( 1 - p ) / p );
  return k;
}*/

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// GammaRandomDraw()
/////////////////////////////////////////////////////////////////////////////
/*static double clModelMath::GammaRandomDraw( const double a, const double b )
{
  // assume a > 0
  unsigned int na = floor( a );

  if ( a == na )
  {
    return b * gsl_ran_gamma_int( na );
  }
  else if ( na == 0 )
  {
    return b * gamma_frac( a );
  }
  else
  {
    return b * ( gsl_ran_gamma_int( na ) + gamma_frac( a - na ) );
  }
} */

/////////////////////////////////////////////////////////////////////////////
// gsl_ran_gamma_int()
/////////////////////////////////////////////////////////////////////////////
/*static clModelMath::double gsl_ran_gamma_int( const unsigned int a )
{
  if ( a < 12 )
  {
    unsigned int i;
    double prod = 1;

    for ( i = 0; i < a; i++ )
    {
      prod *= GetRand();
    }

    // Note: for 12 iterations we are safe against underflow, since
    //the smallest positive random number is O(2^-32). This means the smallest
    //possible product is 2^(-12*32) = 10^-116 which
    //is within the range of double precision.

    return -log( prod );
  }
  else
  {
    return gamma_large( ( double )a );
  }
} */

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// gamma_large()
/////////////////////////////////////////////////////////////////////////////
/*static double clModelMath::gamma_large( const double a )
{
  // Works only if a > 1, and is most efficient if a is large

  //This algorithm, reported in Knuth, is attributed to Ahrens.  A faster one,
  //we are told, can be found in: J. H. Ahrens and U. Dieter, Computing 12
  //(1974) 223-246.

  double sqa, x, y, v;
  sqa = sqrt( 2 * a - 1 );
  do
  {
    do
    {
      y = tan( M_PI * GetRand() );
      x = sqa * y + a - 1;
    }
    while ( x <= 0 );
    v = GetRand();
  }
  while ( v > ( 1 + y * y ) * exp( ( a - 1 ) * log( x / ( a - 1 ) ) - sqa * y ) );

  return x;
} */

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// gamma_frac()
/////////////////////////////////////////////////////////////////////////////
/*static double clModelMath::gamma_frac( const double a )
{
  // This is exercise 16 from Knuth; see page 135, and the solution is on
  //page 551.

  double p, q, x, u, v;
  p = M_E / ( a + M_E );
  do
  {
    u = GetRand();
    v = GetRand();

    if ( u < p )
    {
      x = exp( ( 1 / a ) * log( v ) );
      q = exp( -x );
    }
    else
    {
      x = 1 - log( v );
      q = exp( ( a - 1 ) * log( x ) );
    }
  }
  while ( GetRand() >= q );

  return x;
} */

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// NormalRandomDraw()
/////////////////////////////////////////////////////////////////////////////
float clModelMath::NormalRandomDraw(float fStdDev)
{
  double x, y, r2;

  do
    {
      //choose x,y in uniform square (-1,-1) to (+1,+1)

      x = -1 + 2 * GetRand();
      y = -1 + 2 * GetRand();

      // see if it is in the unit circle
      r2 = x * x + y * y;
    }
  while (r2 > 1.0 || r2 == 0);

  // Box-Muller transform - only use one of the deviates returned
  return fStdDev * y * sqrt (-2.0 * log (r2) / r2);
}
//---------------------------------------------------------------------------
/////////////////////////////////////////////////////////////////////////////
// AddBarkToDBH()
/////////////////////////////////////////////////////////////////////////////
float clModelMath::AddBarkToDBH(float fDIB, float fA, float fB, float fC) {
  if (fDIB <= 0) {
    modelErr stcErr;
    strcpy( stcErr.cFunction, "clModelMath::AddBarkToDBH" );
    strcpy( stcErr.cMoreInfo, "DBH is not a positive number." );
    stcErr.iErrorCode = BAD_DATA;
    throw( stcErr );
  }
  if (fDIB >= 1E18) {
    modelErr stcErr;
    strcpy( stcErr.cFunction, "clModelMath::AddBarkToDBH" );
    strcpy( stcErr.cMoreInfo, "DBH is too large." );
    stcErr.iErrorCode = BAD_DATA;
    throw( stcErr );
  }

  return (fA + (fB * fDIB) + (fC * (fDIB * fDIB)));
}

/*
/////////////////////////////////////////////////////////////////////////////
//     NormalRandomDraw()
/////////////////////////////////////////////////////////////////////////////
float clModelMath::NormalRandomDraw(float fStdDev)
{
  //Returns a normally distributed deviate with zero mean and unit variance
  static int iset = 0;
  static float gset;
  float fac, rsq, v1, v2;

  if ( iset == 0 )
  {
    //We don't have an extra deviate handly, so calculate a new set
    do
    {
      //Pick two uniform numbers in the square extending from -1 to +1 in each direction
      v1 = 2.0 * GetRand() - 1.0;
      v2 = 2.0 * GetRand() - 1.0;
      //See if they are in the unit circle
      rsq = v1 * v1 + v2 * v2;
      //And if they're not, try again
    }
    while ( rsq >= 1.0 || rsq == 0.0 );
    fac = sqrt( -2.0*log( rsq ) / rsq );
    //Now make the Box-Muller transformation to get two normal deviates.
    //Return one and save the other for next time.
    gset = v1 * fac;
    //Set the flag indicating that there is another deviate waiting
    iset = 1;
    return v2 * fac;
  }
  else
  {
    //We have an extra deviate handy, so unset the flag, and return it
    iset = 0;
    return gset;
  }
}
//---------------------------------------------------------------------------
*/
