//---------------------------------------------------------------------------
#include <iostream.h>
#include <stdlib.h>
#include <time.h>
#include <fstream.h>
#include <sstream.h>

#include "SimManager.h"
#include "Plot.h"
#include "BehaviorBase.h"
#include "TreePopulation.h"
#include "Output.h"
#include "XercesClasses.h"
#include "Messages.h"
#include "Behaviors.h"
#include "Constants.h"
#include "Populations.h"
#include "TheGrids.h"
#include "GridBase.h"
#include "PopulationBase.h"
#include "ParsingFunctions.h"
#include "ModelMath.h"

//XML includes
#include <xercesc/dom/DOMBuilder.hpp>
#include <xercesc/util/PlatformUtils.hpp>
//////////////////////////////////////////////////////////////////////////////
// Constructor()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
//////////////////////////////////////////////////////////////////////////////
clSimManager::clSimManager( int iMajorVersion, int iMinorVersion, char *cAppPath)
{
  try
  {
    m_iMajorVersion = iMajorVersion;
    m_iMinorVersion = iMinorVersion;
    strcpy(m_cAppPath, cAppPath);

    mp_oBehaviorManager = NULL;
    mp_oPopulationManager = NULL;
    mp_oGridManager = NULL;
    mp_oPlot = NULL;
    mp_commands = NULL;
    m_iNumCommands = 0;

    m_eSimState = No_Data;

    //Prepare XML parsing tools
    //I must confess I don't exactly know what's going on here - I'm following
    //the Xerces examples
    XMLPlatformUtils::Initialize();
    DOMImplementation * p_oImpl = DOMImplementationRegistry::getDOMImplementation( XMLString::transcode( "LS" ) );
    mp_oXMLParser = ( ( DOMImplementationLS * ) p_oImpl )->createDOMBuilder( DOMImplementationLS::MODE_SYNCHRONOUS, 0 );
    mp_oXMLErrorHandler = new clXercesErrorHandler;
    mp_oXMLParser->setErrorHandler( mp_oXMLErrorHandler );

    GoToNoDataState();
  }
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::clSimManager" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// Destructor()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// May 19, 2004 - Adding the NULLing of pointers (LEM)
////////////////////////////////////////////////////////////////////////////*/
clSimManager::~clSimManager()
{
  if ( No_Data != m_eSimState )
    GoToNoDataState();

  delete mp_oPopulationManager; mp_oPopulationManager = NULL;
  delete mp_oGridManager; mp_oGridManager = NULL;
  delete mp_oBehaviorManager; mp_oBehaviorManager = NULL;
  delete mp_oPlot; mp_oPlot = NULL;

  //Shut down the XML tools
  delete mp_oXMLErrorHandler; mp_oXMLErrorHandler = NULL;
  delete mp_oXMLParser; mp_oXMLParser = NULL;
  XMLPlatformUtils::Terminate();
}

/////////////////////////////////////////////////////////////////////////////
// GoToNoDataState()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clSimManager::GoToNoDataState()
{
  try
  {
    //Create the object managers, if they haven't been already
    if ( !mp_oPopulationManager )
      mp_oPopulationManager = new clPopulationManager( this );
    if ( !mp_oGridManager )
      mp_oGridManager = new clGridManager( this );
    if ( !mp_oBehaviorManager )
      mp_oBehaviorManager = new clBehaviorManager( this );

    //Set up the plot, if it hasn't been already
    if ( !mp_oPlot )
      mp_oPlot = new clPlot( this );

    //Have the object managers and plot clear their memory, in case we're coming
    //from a data defined state
    if ( m_eSimState > No_Data )
    {
      mp_oGridManager->FreeMemory();
      mp_oPopulationManager->FreeMemory();
      mp_oBehaviorManager->FreeMemory();
      delete mp_oPlot;
      mp_oPlot = new clPlot( this );
    }

    //Re-initialize variables that come from input files
    m_iNumTimesteps = 0;
    m_iRandomSeed = 0;

    m_iCurrentTimestep = 0;
    m_iTargetTimestep = 0;

    //Delete the commands
    delete[] mp_commands; mp_commands = NULL;
    m_iNumCommands = 0;

    m_eSimState = No_Data;
  } //end of try block
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::GoToNoDataState" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// ReadFile()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clSimManager::ReadFile( char * cFileName )
{
  try
  {
    DOMDocument * p_oDoc;
    fileType iFileType;

    //Check the file type
    iFileType = GetFileType( cFileName );

    //Parameter file - load it
    if ( parfile == iFileType )
    {
      p_oDoc = mp_oXMLParser->parseURI( cFileName );
      strncpy( m_cParFilename, cFileName, MAX_FILENAME_SIZE - 1 );
      //This doesn't always happen with strncpy - make it explicit
      m_cParFilename[MAX_FILENAME_SIZE - 1] = '\0';
      ReadParameterFile( p_oDoc, iFileType );

      //Batch file - erase memory and run the batch
    }
    else if ( batchfile == iFileType )
    {
      GoToNoDataState();
      RunBatch( cFileName );

      //Detailed output file - load it like a par file
    }
    else if ( detailed_output == iFileType )
    {
      p_oDoc = mp_oXMLParser->parseURI( cFileName );
      strncpy( m_cParFilename, cFileName, MAX_FILENAME_SIZE - 1 );
      //This doesn't always happen with strncpy - make it explicit
      m_cParFilename[MAX_FILENAME_SIZE - 1] = '\0';
      ReadParameterFile( p_oDoc, iFileType );

      //Tree file - pass it to the tree population
    }
    else if ( tree == iFileType )
    {
      //This is only allowed if a parameter file is loaded (i.e. the model is
      //ready to run)
      if ( Initialized != m_eSimState )
      {
        modelErr stcErr;
        stcErr.iErrorCode = MODEL_NOT_READY;
        strcpy( stcErr.cFunction, "clSimManager::ReadFile" );
        strcpy( stcErr.cMoreInfo, "Parameter file must be loaded" );
        throw( stcErr );
      }
      else
      {
        p_oDoc = mp_oXMLParser->parseURI( cFileName );
        ReadTreeFile( p_oDoc, iFileType );
      }

      //Tree map file - pass it to the tree population
    }
    else if ( treemap == iFileType )
    {
      //This is only allowed if a parameter file is loaded (i.e. the model is
      //ready to run)
      if ( Initialized != m_eSimState )
      {
        modelErr stcErr;
        stcErr.iErrorCode = MODEL_NOT_READY;
        strcpy( stcErr.cFunction, "clSimManager::ReadFile" );
        strcpy( stcErr.cMoreInfo, "Parameter file must be loaded" );
        throw( stcErr );
      }
      else
      {
        p_oDoc = mp_oXMLParser->parseURI( cFileName );
        ReadTreeFile( p_oDoc, iFileType );
      }

      //Grid map file - pass it to the grid manager
    }
    else if ( map == iFileType )
    {
      //This is only allowed if a parameter file is loaded (i.e. the model is
      //ready to run)
      if ( Initialized != m_eSimState )
      {
        modelErr stcErr;
        stcErr.iErrorCode = MODEL_NOT_READY;
        strcpy( stcErr.cFunction, "clSimManager::ReadFile" );
        strcpy( stcErr.cMoreInfo, "Parameter file must be loaded" );
        throw( stcErr );
      }
      else
      {
        p_oDoc = mp_oXMLParser->parseURI( cFileName );
        ReadMapFile( p_oDoc, iFileType );
      }

      //Detailed output timestep file - read it like tree and grid maps
    }
    else if ( detailed_output_timestep == iFileType )
    {
      //This is only allowed if a parameter file is loaded (i.e. the model is
      //ready to run)
      if ( Initialized != m_eSimState )
      {
        modelErr stcErr;
        stcErr.iErrorCode = MODEL_NOT_READY;
        strcpy( stcErr.cFunction, "clSimManager::ReadFile" );
        strcpy( stcErr.cMoreInfo, "Parameter file must be loaded" );
        throw( stcErr );
      }
      else
      {
        p_oDoc = mp_oXMLParser->parseURI( cFileName );
        ReadMapFile( p_oDoc, iFileType );
        ReadTreeFile( p_oDoc, iFileType );
      }
    }
    else
    {
      //This is not a type we can handle - throw an error
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE_TYPE;
      strcpy( stcErr.cFunction, "clSimManager::ReadFile" );
      strcpy( stcErr.cMoreInfo, cFileName );
      throw( stcErr );
    }

    //Free the DOM document's memory
    mp_oXMLParser->resetDocumentPool();
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::ReadFile" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// GetFileType()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
fileType clSimManager::GetFileType( char * cFileName )
{
  try
  {
    string strFileExtension, strFileNameCopy, strPlaybackExtension, strFileCodeTag = "fileCode", //the XML tag for the SORTIE file
         //code
         strLine, //for working with lines of a text file after converting
         //them from char*
         strFileCode; //for holding the text of the file code from an XML file
    string::size_type pos; //for string parsing - holds string positions
    fstream file;
    FILE * filetest; //only this kind of file can be tested for existence of a
    //file without creating
    char cXMLTest[5], //for extracting the first 4 characters of a text file
         cLine[200], //for extracting text from file and converting to string
         cXMLDec[] = "<?xm", cXMLCom[] = "<!--";
    int iFileType = 0; //for an int representation of the file type code
    modelErr stcErr;
    strcpy( stcErr.cFunction, "clSimManager::GetFileType" );

    //Does this file exist and can it be opened?
    filetest = fopen( cFileName, "r" );
    if ( filetest == NULL )
    { //file does not exist
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cMoreInfo, cFileName );
      throw( stcErr );
    }
    else
      fclose( filetest );

    strFileNameCopy = cFileName;

    //Find this file's extension
    pos = strFileNameCopy.find_last_of( "." );
    if ( pos == string::npos ) //no extension - return "not sortie"
           return notrecognized;

    strFileExtension = strFileNameCopy.substr( pos, strFileNameCopy.length() );

    /** IS THIS XML? */
    //Look at the first four characters of the file - should be either <!--
    //(a comment) or <?xm (the xml declaration)
    file.open( cFileName, ios::in );
    file.get( cXMLTest, sizeof( cXMLTest ) );
    strcat( cXMLTest, "\0" );
    if ( strcmp( cXMLTest, cXMLDec ) == 0 || strcmp( cXMLTest, cXMLCom ) == 0 )
    {
      //Yes - this is an xml file - so search the file text for the fileCode tag
      while ( 1 )
      { //read all the file text until we find what we want
        //search the line from the file for the file code tag
        file.getline( cLine, sizeof( cLine ) );
        strLine = cLine;
        pos = strLine.find( strFileCodeTag );
        if ( pos != string::npos )
        {
          //we found the file code tag - now extract the 8 char code after it
          strFileCode = strLine.substr( pos + strFileCodeTag.length() + 2, 8 );
          //the file type is the third pair of numbers
          strFileCode = strFileCode.substr( 4, 2 );
          iFileType = atoi( strFileCode.c_str() );
          break;
        }
        if ( file.eof() ) break;
      }
      //if we don't have a file code this isn't an XML file we can read - exit
      if ( 0 == iFileType ) return notrecognized;

      //this is an XML file and we have a file type code - see if it's defined
      if ( iFileType > 0 && iFileType < lastfile )
        return iFileType; //yep - it was found
      else
        return notrecognized; //we couldn't tell what file type it was
    } //end of if (strcmp(cXMLTest, "<?xm") == 0 || strcmp(cXMLTest, "<!--") == 0)

    /** IS THIS OLD SORTIE? This isn't an XML file so see if it's an old SORTIE text file */
    //Check the file extension
    if ( strFileExtension == ".sgli" || strFileExtension == ".par" || strFileExtension == ".out"
         || strFileExtension == ".stmf" || strFileExtension == ".sll" || strFileExtension == ".hvr"
         || strFileExtension == ".spb" || strFileExtension == ".hvs" ) return oldsortie;

    return notrecognized;
  }
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::GetFileType" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------
fileType clSimManager::GetFileType( DOMDocument * p_oDoc )
{
  try
  {
    DOMElement * p_oRoot; //for getting the root element
    DOMAttr * p_oAttribute; //for getting the filecode node
    std::string strData; //for extracting the filecode text
    int iFileType; //for storing the extracted file type

    p_oRoot = p_oDoc->getDocumentElement();
    p_oAttribute = p_oRoot->getAttributeNode( XMLString::transcode( "fileCode" ) );
    if ( NULL == p_oAttribute ) return notrecognized;
    const XMLCh * test = p_oAttribute->getValue();
    strData = XMLString::transcode( test );
    //strData = XMLString::transcode(p_oAttribute->getValue());
    strData = strData.substr( 4, 2 );
    iFileType = atoi( strData.c_str() );

    //if we don't have a file code this isn't an XML file we can read - exit
    if ( 0 == iFileType ) return notrecognized;

    //See if the file type code is defined
    if ( iFileType > 0 && iFileType < lastfile )
      return iFileType; //yep - it was found
    else
      return notrecognized; //we couldn't tell what file type it was
  }
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::GetFileType" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// RunSim()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// November 8, 2004 - Added a check for a quit message (LEM)
////////////////////////////////////////////////////////////////////////////*/
unsigned long clSimManager::RunSim( int iNumStepsToRun )
{
  try
  {
    DOMDocument * p_oDoc; //in case we need to parse and reload the par file
    clWorkerBase * p_oTempObject = NULL;
    clBehaviorBase * p_oBehavior = NULL;
    clPopulationBase * p_oPopulation = NULL;
    modelMsg stcMsg; //for checking for messages
    unsigned int iNumBehaviors = 0, //number of behaviors - for looping through
         //and calling their Action() functions
         iNumPopulations = 0, //number of populations
         iTimestep = 0, //current timestep - loop counter
         iStartTimestep = 0, //the timestep to start on this time
         iEndTimestep = 0; //the timestep to end on this time
    clock_t startRun, endRun;
    unsigned long iRunTime;

    m_bUserQuit = false;

    //Make sure that the number of timesteps is not less than zero
    if ( 0 > iNumStepsToRun )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::RunSim" );
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cMoreInfo, "Number of timesteps to run must be greater than zero." );
      throw( stcErr );
    }

    //If the status is "Run_Complete" - reset the model
    if ( Run_Complete == m_eSimState )
    {
      p_oDoc = mp_oXMLParser->parseURI( m_cParFilename );
      ReadParameterFile( p_oDoc, GetFileType( p_oDoc ) );
      //Free DOM doc memory
      mp_oXMLParser->resetDocumentPool();
    }

    //Only continue if the Simulation Manager state is Initialized or Paused
    if ( m_eSimState != Initialized && m_eSimState != Paused )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::RunSim" );
      stcErr.iErrorCode = MODEL_NOT_READY;
      strcpy( stcErr.cMoreInfo, "Valid parameter file not loaded" );
      throw( stcErr );
    }

    startRun = clock();


    //If the model status is paused and no timesteps are entered, pick up where
    //we left off
    if ( Paused == m_eSimState && 0 == iNumStepsToRun )
    {
      //If the current timestep is the same as the target timestep (minus 1),
      //run to the end
      if ( m_iCurrentTimestep == m_iTargetTimestep - 1 )
      {
        m_iTargetTimestep = 0;
        iEndTimestep = m_iNumTimesteps + 1;
      }
      else
      {
        iEndTimestep = m_iTargetTimestep;
      }
      iStartTimestep = m_iCurrentTimestep + 1;
      m_eSimState = Initialized; //no longer paused - running again
    } //end of if (Paused == m_eSimState && 0 != iNumStepsToRun)

    //Not paused, or we are supposed to start over - so figure out the start and
    //stop timesteps
    else
    {
      m_eSimState = Initialized; //erase paused status if present
      iStartTimestep = m_iCurrentTimestep + 1;
      if ( iNumStepsToRun < 1 || iNumStepsToRun >= ( m_iNumTimesteps - m_iCurrentTimestep ) )
        iEndTimestep = m_iNumTimesteps + 1;
      else
        iEndTimestep = iStartTimestep + iNumStepsToRun;

      m_iTargetTimestep = iEndTimestep;
    } //end of else - not paused

    //Query the behavior manager for the number of behaviors to run
    //each timestep
    iNumBehaviors = mp_oBehaviorManager->GetNumberOfObjects();
    iNumPopulations = mp_oPopulationManager->GetNumberOfObjects();

    //TIMESTEP LOOP
    for ( iTimestep = iStartTimestep; iTimestep < iEndTimestep; iTimestep++ )
    {

      //Pass a message indicating the timestep we just started
      stcMsg.iMessageCode = INFO;
      stringstream msg;
      msg << "Starting timestep " << iTimestep << " of " << m_iNumTimesteps;
      strcpy( stcMsg.cMoreInfo, msg.str().c_str() );
      SendMessage( stcMsg );

      m_iCurrentTimestep = iTimestep;

      //Run each behavior object
      for ( int i = 0; i < iNumBehaviors; i++ )
      {
        p_oTempObject = mp_oBehaviorManager->PassObjectPointer( i );
        p_oBehavior = dynamic_cast < clBehaviorBase * > ( p_oTempObject );
        p_oBehavior->Action();
        for ( int j = 0; j < iNumPopulations; j++ )
        {
          p_oTempObject = mp_oPopulationManager->PassObjectPointer( j );
          p_oPopulation = dynamic_cast < clPopulationBase * > ( p_oTempObject );
          p_oPopulation->DoDataUpdates();
        }
      }

      //Run the cleanup operations
      TimestepCleanup();

      //Check to see if we need to process any messages while we were waiting
      stcMsg = CheckForMessage(m_cAppPath);
      if ( NO_MESSAGE != stcMsg.iMessageCode )
      {
        if ( MODEL_PAUSED == stcMsg.iMessageCode )
        {
          //"pause"; set the timestep we were interrupted on so we know where
          //to pick up
          m_eSimState = Paused;
          m_iTargetTimestep = iEndTimestep;
          stcMsg.iMessageCode = MODEL_PAUSED;
          sprintf(stcMsg.cMoreInfo, "%s%d", "Finished timestep ", m_iCurrentTimestep);
          SendMessage( stcMsg );
          return 0;
        }
        else if ( QUIT == stcMsg.iMessageCode )
        {
          //quit - shut down everything
          GoToNoDataState();
          m_bUserQuit = true;
          return 0;
        }
      }
    } //end of timestep loop

    //Set the sim state to run complete if we've run all the timesteps
    if ( m_iCurrentTimestep == m_iNumTimesteps )
    {
      //Perform end-of-run cleanup
      EndOfRunCleanup();
      m_eSimState = Run_Complete;
      m_iTargetTimestep = 0;
    }
    //Otherwise the status is paused
    else {
      m_eSimState = Paused;
      stcMsg.iMessageCode = MODEL_PAUSED;
      SendMessage( stcMsg );
    }

    endRun = clock();
    iRunTime = endRun - startRun;
    return ( iRunTime / CLK_TCK ); //convert to seconds

  } //end of try block
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::RunSim" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// TimestepCleanup()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clSimManager::TimestepCleanup()
{
  try
  {
    //Run the timestep cleanup functions for each of the object managers
    //and the plot
    mp_oPlot->TimestepCleanup();
    mp_oBehaviorManager->TimestepCleanup();
    mp_oGridManager->TimestepCleanup();
    mp_oPopulationManager->TimestepCleanup();
  } //end of try block
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::TimestepCleanup" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// EndOfRunCleanup()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clSimManager::EndOfRunCleanup()
{
  try
  {
    //Run the end-of-run cleanup functions for each of the object managers
    //and the plot
    mp_oPlot->EndOfRunCleanup();
    mp_oBehaviorManager->EndOfRunCleanup();
    mp_oGridManager->EndOfRunCleanup();
    mp_oPopulationManager->EndOfRunCleanup();
  } //end of try block
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::EndOfRunCleanup" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// GetPopulationObject()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
clPopulationBase * clSimManager::GetPopulationObject( char * cPopName )
{
  try
  {
    clWorkerBase * p_oTempObject = NULL;
    clPopulationBase * p_oPop = NULL;

    //Make sure that the population manager has been created - if not, just
    //return NULL
    if ( !mp_oPopulationManager ) return p_oPop;

    //Request the population from the Population Manager and return it
    p_oTempObject = mp_oPopulationManager->PassObjectPointer( cPopName );
    //Recast the pointer as a population pointer
    if ( p_oTempObject )
      p_oPop = dynamic_cast < clPopulationBase * > ( p_oTempObject );

    return p_oPop;
  }
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::GetPopulationObject" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------
clPopulationBase * clSimManager::GetPopulationObject( int iIndex )
{
  try
  {
    clWorkerBase * p_oTempObject = NULL;
    clPopulationBase * p_oPop = NULL;

    //Make sure that the population manager has been created - if not, just
    //return NULL
    if ( !mp_oPopulationManager ) return p_oPop;

    //Request the population from the Population Manager and return it
    p_oTempObject = mp_oPopulationManager->PassObjectPointer( iIndex );
    //Recast the pointer as a population pointer
    if ( p_oTempObject )
      p_oPop = dynamic_cast < clPopulationBase * > ( p_oTempObject );

    return p_oPop;
  }
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::GetPopulationObject" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// GetGridObject()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
clGridBase * clSimManager::GetGridObject( char * cGridName )
{
  try
  {
    clWorkerBase * p_oTempObject = NULL;
    clGridBase * p_oGrid = NULL;

    //Make sure the Grid Manager has been created - if not, just return NULL
    if ( !mp_oGridManager ) return p_oGrid;

    //Request the grid from the Grid Manager and return it
    p_oTempObject = mp_oGridManager->PassObjectPointer( cGridName );
    //Recast the pointer as a population pointer
    if ( p_oTempObject )
      p_oGrid = ( clGridBase * ) p_oTempObject;

    return p_oGrid;
  }
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::GetGridObject" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------
clGridBase * clSimManager::GetGridObject( int iIndex )
{
  try
  {
    clWorkerBase * p_oTempObject = NULL;
    clGridBase * p_oGrid = NULL;

    //Make sure the Grid Manager has been created - if not, just return NULL
    if ( !mp_oGridManager ) return p_oGrid;

    //Request the grid from the Grid Manager and return it
    p_oTempObject = mp_oGridManager->PassObjectPointer( iIndex );
    //Recast the pointer as a population pointer
    if ( p_oTempObject )
      p_oGrid = ( clGridBase * ) p_oTempObject;

    return p_oGrid;
  }
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::GetGridObject" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// GetBehaviorObject()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
clBehaviorBase * clSimManager::GetBehaviorObject( char * cBehaviorName )
{
  try
  {
    clWorkerBase * p_oTempObject = NULL;
    clBehaviorBase * p_oBehavior = NULL;

    //Make sure the Behavior Manager has been created - if not, just return NULL
    if ( !mp_oBehaviorManager ) return p_oBehavior;

    //Request the behavior from the Behavior Manager and return it
    p_oTempObject = mp_oBehaviorManager->PassObjectPointer( cBehaviorName );
    //Recast the pointer as a population pointer
    if ( p_oTempObject )
      p_oBehavior = dynamic_cast < clBehaviorBase * > ( p_oTempObject );

    return p_oBehavior;
  }
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::GetBehaviorObject" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------
clBehaviorBase * clSimManager::GetBehaviorObject( int iIndex )
{
  try
  {
    clWorkerBase * p_oTempObject = NULL;
    clBehaviorBase * p_oBehavior = NULL;

    //Make sure the Behavior Manager has been created - if not, just return NULL
    if ( !mp_oBehaviorManager ) return p_oBehavior;

    //Request the behavior from the Behavior Manager and return it
    p_oTempObject = mp_oBehaviorManager->PassObjectPointer( iIndex );
    //Recast the pointer as a population pointer
    if ( p_oTempObject )
      p_oBehavior = dynamic_cast < clBehaviorBase * > ( p_oTempObject );

    return p_oBehavior;
  }
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::GetGridObject" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// ReadParameterFile()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clSimManager::ReadParameterFile( DOMDocument * p_oDoc, fileType iFileType )
{
  try
  {
    int iNumObjects, //for looping through objects in an object manager
         i; //loop counter

    //Make sure our pointer is good
    if ( NULL == p_oDoc )
    { //nope - throw error
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::ReadParameterFile" );
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cMoreInfo, "Document pointer is NULL." );
      throw( stcErr );
    }

    //If there's currently data read in, erase it all
    if ( No_Data != m_eSimState ) GoToNoDataState();

    //Call the simulation manager's own DoSetup, which will read general params
    DoSetup( p_oDoc );

    //Set up the plot object
    mp_oPlot->DoObjectSetup( p_oDoc, iFileType );

    /** Start by creating the objects */
    //Do populations
    mp_oPopulationManager->CreateObjects( p_oDoc );

    //Do grids
    mp_oGridManager->CreateObjects( p_oDoc );

    //Do behaviors
    mp_oBehaviorManager->CreateObjects( p_oDoc );

    /** Phase 2 setup */
    //Do populations
    mp_oPopulationManager->DoObjectSetup( p_oDoc, iFileType );

    //Do grids
    mp_oGridManager->DoObjectSetup( p_oDoc, iFileType );

    //Do behaviors
    mp_oBehaviorManager->DoObjectSetup( p_oDoc, iFileType );

    //Set the sim state - if we were paused mid-run, keep the status as
    //paused
    if ( Paused != m_eSimState )
      m_eSimState = Initialized;
  } //end of try block
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::ReadParameterFile" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// ReadTreeFile()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clSimManager::ReadTreeFile( DOMDocument * p_oDoc, fileType iFileType )
{
  try
  {
    clWorkerBase * p_oTreePop = NULL;

    //Make sure that we are not in the No_Data state
    if ( No_Data == m_eSimState )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::ReadTreeFile" );
      stcErr.iErrorCode = MODEL_NOT_READY;
      throw( stcErr );
    }

    //Make sure our pointer is good
    if ( NULL == p_oDoc )
    { //nope - throw error
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::ReadTreeFile" );
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cMoreInfo, "Document pointer is NULL." );
      throw( stcErr );
    }

    //Request the tree population from the Population Manager
    p_oTreePop = mp_oPopulationManager->PassObjectPointer( "treepopulation" );

    //Make sure we got a valid pointer - if not throw error
    if ( NULL == p_oTreePop )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::ReadTreeFile" );
      stcErr.iErrorCode = CANT_FIND_OBJECT;
      throw( stcErr );
    }

    //Pass it to the tree population for setup
    p_oTreePop->DoObjectSetup( p_oDoc, iFileType );
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::ReadTreeFile" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// ReadMapFile()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clSimManager::ReadMapFile( DOMDocument * p_oDoc, fileType iFileType )
{
  try
  {
    //Make sure that we are not in the No_Data state
    if ( No_Data == m_eSimState )
    {
      modelErr stcErr;
      stcErr.iErrorCode = MODEL_NOT_READY;
      strcpy( stcErr.cFunction, "clSimManager::ReadMapFile" );
      throw( stcErr );
    }

    //Make sure our pointer is good
    if ( NULL == p_oDoc )
    { //nope - throw error
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::ReadMapFile" );
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cMoreInfo, "Document pointer is NULL." );
      throw( stcErr );
    }

    //Pass it to the grid manager for setup
    mp_oGridManager->DoObjectSetup( p_oDoc, iFileType );
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::ReadMapFile" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// DoSetup()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clSimManager::DoSetup( DOMDocument * p_oDoc )
{
  try
  {
    DOMElement * p_oElement = p_oDoc->getDocumentElement();
    int iTemp;

    //Get number of timesteps
    FillSingleValue( p_oElement, "timesteps", & iTemp, true );
    if ( 0 >= iTemp )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::DoSetup" );
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cMoreInfo, "Timesteps must be a positive number." );
      throw( stcErr );
    }
    m_iNumTimesteps = iTemp;

    //Get number of years per timestep
    FillSingleValue( p_oElement, "yearsPerTimestep", & m_fNumYearsPerTimestep, true );

    //Throw an error if the value is less than 0
    if ( 0 >= m_fNumYearsPerTimestep )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::DoSetup" );
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cMoreInfo, "Number of years per timestep must be a positive number." );
      throw( stcErr );
    }

    //Get current timestep, if present
    iTemp = 0;
    FillSingleValue( p_oElement, "currentTimestep", & iTemp, false );

    //Make sure that the current timestep is not a negative number or
    //greater than the number of timesteps
    if ( iTemp < 0 || iTemp >= m_iNumTimesteps )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::DoSetup" );
      stcErr.iErrorCode = BAD_DATA;
      sprintf( stcErr.cMoreInfo, "%s%d", "Current timestep is not within run - value is ", iTemp );
      throw( stcErr );
    }
    m_iCurrentTimestep = iTemp;

    //Get random number seed
    FillSingleValue( p_oElement, "randomSeed", & m_iRandomSeed, true );
    //Initialize random number generator - if a seed was provided, use it; if
    //the seed was 0, use a random seed
    if ( 0 == m_iRandomSeed )
    {
      time(&m_iActualSeed);
      m_iActualSeed = -m_iActualSeed;
    }
    else if ( m_iRandomSeed > 0 )
    {
      m_iActualSeed = -m_iRandomSeed;
    }
    else
    {
      m_iActualSeed = m_iRandomSeed;
    }

    clModelMath::SetRandomSeed(m_iActualSeed);

    //log.open(logFileName, ios::out | ios::app);
    //log << "Beginning of run for par file " << m_strParFilename << "\n";
    //log.close();

  } //end of the try block
  catch ( modelErr & err )
  {
    GoToNoDataState();
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::DoSetup" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// DoCommand()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
bool clSimManager::DoCommand( char * cCommand, char * cArguments )
{
  try
  {
    bool bFoundCommand = false; //whether or not we found a matching command

    //Check all the registered commands to see if we have a match
    for ( int i = 0; i < m_iNumCommands; i++ )
    {
      if ( strcmp( cCommand, mp_commands[i].cCommand ) == 0 )
      {
        mp_commands[i].p_oObject->DoCommand( cCommand, cArguments );
        bFoundCommand = true;
        break;
      }
    }
    return bFoundCommand;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  { //non-fatal error
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::DoCommand" );
    GoToNoDataState();
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/*/ ///////////////////////////////////////////////////////////////////////////
     WriteCommandHelp()
     ////////////////////////////////////////////////////////////////////////////*/
     /*bool clSimManager::WriteCommandHelp(char *cCommand) { try {
     bool bFoundCommand = false;  //whether or not we found a matching command

     for (int i = 0; i < m_iNumCommands; i++) { if (strcmp(cCommand, mp_commands[i].cCommand) == 0) {
     mp_commands[i].p_oObject->WriteCommandHelp(cCommand); bFoundCommand = true; break; } } return bFoundCommand;
     } //end of try block catch (modelErr &err) { throw(err); } catch (modelMsg &msg) { //non-fatal error throw(msg); }
     catch (...) { modelErr stcErr; stcErr.iErrorCode = UNKNOWN; strcpy(stcErr.cFunction, "clSimManager::WriteCommandHelp");
     GoToNoDataState(); throw(stcErr); } }  */
     //----------------------------------------------------------------------------


     /////////////////////////////////////////////////////////////////////////////
     // RegisterCommand()
     // Edit history:
     // -------------
     // April 28, 2004 - Submitted in beta version (LEM)
     /////////////////////////////////////////////////////////////////////////////
     void clSimManager::RegisterCommand( char * cCommand, clWorkerBase * p_oObject )
     {
       try
       {
         stcCommandList * p_listCopy; //for making a copy of the command list
         short int iIndex = -1, //array index
              i; //loop counter

         //If there are no commands registered so far, register this one
         if ( NULL == mp_commands )
         {
           m_iNumCommands = 1;
           mp_commands = new stcCommandList[m_iNumCommands];
           strcpy( mp_commands[0].cCommand, cCommand );
           mp_commands[0].p_oObject = p_oObject;
         } //end of if (NULL == mp_commands)
         else
         {

           //First check to see if this is a duplicate
           for ( i = 0; i < m_iNumCommands; i++ )
             if ( strcmp( cCommand, mp_commands[i].cCommand ) == 0 ) iIndex = i;

           if ( -1 == iIndex )
           { //not a duplicate
             //Register this command - copy the current list and then increase the
             //list by one
             p_listCopy = new stcCommandList[m_iNumCommands];
             for ( i = 0; i < m_iNumCommands; i++ )
             {
               strcpy( p_listCopy[i].cCommand, mp_commands[i].cCommand );
               p_listCopy[i].p_oObject = mp_commands[i].p_oObject;
             }

             delete[] mp_commands; mp_commands = NULL;
             m_iNumCommands++;
             mp_commands = new stcCommandList[m_iNumCommands];
             for ( int i = 0; i < m_iNumCommands - 1; i++ )
             {
               strcpy( mp_commands[i].cCommand, p_listCopy[i].cCommand );
               mp_commands[i].p_oObject = p_listCopy[i].p_oObject;
             }

             strcpy( mp_commands[m_iNumCommands - 1].cCommand, cCommand );
             mp_commands[m_iNumCommands - 1].p_oObject = p_oObject;

             //delete the list copy
             delete[] p_listCopy;
           } //end of if (-1 == iIndex)
           else
           {
             //This is a duplicate - overwrite the old command with the new one
             mp_commands[iIndex].p_oObject = p_oObject;
           }
         } //end of else
       } //end of try block
       catch ( modelErr & err )
       {
         GoToNoDataState();
         throw( err );
       }
       catch ( modelMsg & msg )
       { //non-fatal error
         throw( msg );
       }
       catch ( ... )
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy( stcErr.cFunction, "clSimManager::Register" );
         GoToNoDataState();
         throw( stcErr );
       }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// RunBatch()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clSimManager::RunBatch( char * cBatchFile )
{
  DOMDocument * p_oBatchFileDoc, //Document object model of the batch file
       * p_oParFileDoc; //DOM of par file
  DOMElement * p_oBatchElement; //Element variable for batch group
  DOMNode * p_oBatchNode, //Node variable for batch group
       * p_oFilenameNode; //Node for detailed output filename
  DOMNodeList * p_oBatchGroups, //List of all batch group
       * p_oNodeList;
  fileType iFileType;



  //This is the list of parameter file names and number of times to run
  //each
  struct stcParFiles
  {
    char cParFileName[MAX_FILENAME_SIZE];
    short int iNumRuns;
  }
  * p_parFileList;

  fstream file; //for checking the validity of file names
  stringstream strNewFileName; //for assembling the new output filenames
  char * cNumRunsText; //text for number of times to run
  std::string strDetailedOutputFileName, //file name of detailed output file, if any
       strShortOutFileName; //file name of short output file, if any
  int iNumBatchGroups = 0, //number of par file / run number pairs
       iNumNodes = 0, //used for counting search results
       i, j; //loop counter

  //***********************************
  //Validation
  //***********************************
  //Make sure the filename passed is not empty
  if ( strlen( cBatchFile ) == 0 )
  {
    modelErr stcErr;
    strcpy( stcErr.cFunction, "clSimManager::RunBatch" );
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cMoreInfo, "Need batch file path." );
    throw( stcErr );
  }

  //Make sure the file is a batch file - if not throw error
  if ( batchfile != GetFileType( cBatchFile ) )
  {
    modelErr stcErr;
    strcpy( stcErr.cFunction, "clSimManager::RunBatch" );
    stcErr.iErrorCode = BAD_FILE_TYPE;
    throw( stcErr );
  }

  //***********************************
  //Parse the batch document, and put its info into the filenames array.  Each
  //parameter file has an array bucket, with its associated number of runs to
  //complete
  //***********************************
  //Parse the document
  p_oBatchFileDoc = mp_oXMLParser->parseURI( cBatchFile );

  //Count the number of batch groups
  p_oBatchGroups = p_oBatchFileDoc->getElementsByTagName( XMLString::transcode( "ba_parFile" ) );
  iNumBatchGroups = p_oBatchGroups->getLength();

  //Declare the filename array to match
  p_parFileList = new struct stcParFiles[iNumBatchGroups];

  for ( i = 0; i < iNumBatchGroups; i++ )
  {
    //Cast the batch group node to an element for extracting the data
    p_oBatchNode = p_oBatchGroups->item( i );
    p_oBatchElement = ( DOMElement * ) p_oBatchNode;


    //Get the par file name
    p_oNodeList = p_oBatchElement->getElementsByTagName( XMLString::transcode( "ba_fileName" ) );
    iNumNodes = p_oNodeList->getLength();
    if ( 0 == iNumNodes )
    { //no par filename - throw error
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::RunBatch" );
      stcErr.iErrorCode = DATA_MISSING;
      sprintf( stcErr.cMoreInfo, "%s%d", "ba_parFile for batch group ", i );
      throw( stcErr );
    } //end of if (0 == iNumNodes)


    //Assign the par filename into its bucket
    strncpy( p_parFileList[i].cParFileName, XMLString::transcode( p_oNodeList->item( 0 )->getFirstChild()->getNodeValue() ), MAX_FILENAME_SIZE );


    //Check to see if we can open the par file
    file.open( p_parFileList[i].cParFileName );
    if ( file.bad() )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::RunBatch" );
      stcErr.iErrorCode = BAD_FILE;
      sprintf( stcErr.cMoreInfo, "%s%s", "Couldn't open parameter file ", p_parFileList[i].cParFileName );
      throw( stcErr );
    } //end of if (file.bad())
    file.close();

    //Get the number of times to run the par file
    p_oNodeList = p_oBatchElement->getElementsByTagName( XMLString::transcode( "ba_numTimesToRun" ) );
    if ( 0 == p_oNodeList->getLength() )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::RunBatch" );
      stcErr.iErrorCode = DATA_MISSING;
      sprintf( stcErr.cMoreInfo, "%s%s", "ba_numTimesToRun for parameter file ", p_parFileList[i].cParFileName );
      throw( stcErr );
    } //end of if (0 == p_oNodeList->getLength())
    cNumRunsText = XMLString::transcode( p_oNodeList->item( 0 )->getFirstChild()->getNodeValue() );

    //Assign number of runs to its place in the array
    p_parFileList[i].iNumRuns = atoi( cNumRunsText );

    //Make sure the number of runs was a valid, positive integer
    if ( 0 == p_parFileList[i].iNumRuns && strcmp( cNumRunsText, "0" ) != 0 )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::RunBatch" );
      stcErr.iErrorCode = BAD_DATA;
      sprintf( stcErr.cMoreInfo, "%s%s", "Number of runs is not a number:  ", cNumRunsText );
      throw( stcErr );
    }
    else if ( 0 > p_parFileList[i].iNumRuns )
    {
      modelErr stcErr;
      strcpy( stcErr.cFunction, "clSimManager::RunBatch" );
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cMoreInfo, "Number of runs cannot be less than zero." );
      throw( stcErr );
    }
  } //end of for (i = 0; i < iNumBatchGroups; i++)


  //****************************************
  //Main Loop
  //Go through each batch group in the array and run each par file for the
  //requested number of times
  //****************************************
  for ( i = 0; i < iNumBatchGroups; i++ )
  {
    try
    {

      //Set the filename as the current parameter filename
      strncpy( m_cParFilename, p_parFileList[i].cParFileName, MAX_FILENAME_SIZE - 1 );
      //This doesn't always happen with strncpy - make it explicit
      m_cParFilename[MAX_FILENAME_SIZE - 1] = '\0';

      iFileType = GetFileType( m_cParFilename );

      p_oParFileDoc = mp_oXMLParser->parseURI( p_parFileList[i].cParFileName );
      //In batch mode - we need to check for the presence of output settings.
      //If it exists, grab the file names - we'll rename them for each run
      //so they're not overwritten
      p_oNodeList = p_oParFileDoc->getElementsByTagName( XMLString::transcode( "ou_filename" ) );
      if ( 0 != p_oNodeList->getLength() )
      {
        strDetailedOutputFileName = XMLString::transcode( p_oNodeList->item( 0 )->getFirstChild()->getNodeValue() );
        //trim off the file extension, if it exists
        if ( "" != strDetailedOutputFileName )
        {
          strDetailedOutputFileName = strDetailedOutputFileName.substr( 0, strDetailedOutputFileName.find( DETAILED_OUTPUT_FILE_EXT ) );
        }
      }
      else
        strDetailedOutputFileName = "";
      //Short output
      p_oNodeList = p_oParFileDoc->getElementsByTagName( XMLString::transcode( "so_filename" ) );
      if ( 0 != p_oNodeList->getLength() )
      {
        strShortOutFileName = XMLString::transcode( p_oNodeList->item( 0 )->getFirstChild()->getNodeValue() );
        //trim off the file extension, if it exists
        if ( "" != strShortOutFileName )
        {
          strShortOutFileName = strShortOutFileName.substr( 0, strShortOutFileName.find( SHORT_OUTPUT_FILE_EXT ) );
        }
      }
      else
        strShortOutFileName = "";

      //Perform the number of runs specified
      cout << "\nBeginning runs for file " << p_parFileList[i].cParFileName << ".\n" << flush;
      for ( j = 0; j < p_parFileList[i].iNumRuns; j++ )
      {
        //Clean the memory
        GoToNoDataState();

        //Re-parse the parameter file if this isn't the first timestep -
        //otherwise we can just keep the parsed doc from above
        if ( j > 0 )
          p_oParFileDoc = mp_oXMLParser->parseURI( p_parFileList[i].cParFileName );

        //Rename the output files in the DOM
        if ( strDetailedOutputFileName != "" )
        {
          strNewFileName.str( "" );
          strNewFileName << strDetailedOutputFileName << "_" << j + 1;
          p_oNodeList = p_oParFileDoc->getElementsByTagName( XMLString::transcode( "ou_filename" ) );
          p_oFilenameNode = p_oNodeList->item( 0 )->getFirstChild();
          p_oFilenameNode->setNodeValue( XMLString::transcode( strNewFileName.str().c_str() ) );
        }
        if ( strShortOutFileName != "" )
        {
          strNewFileName.str( "" );
          strNewFileName << strShortOutFileName << "_" << j + 1;
          p_oNodeList = p_oParFileDoc->getElementsByTagName( XMLString::transcode( "so_filename" ) );
          p_oFilenameNode = p_oNodeList->item( 0 )->getFirstChild();
          p_oFilenameNode->setNodeValue( XMLString::transcode( strNewFileName.str().c_str() ) );
        }

        //Send in the par file
        ReadParameterFile( p_oParFileDoc, iFileType );

        //Run
        RunSim();

        //If the user quit this run, then quit the batch
        if (m_bUserQuit) {
          return;
        }
        cout << "Run " << ( j + 1 ) << " of " << p_parFileList[i].iNumRuns << " completed.\n" << flush;

        //Free the DOM document's memory
        mp_oXMLParser->resetDocumentPool();
      } //end of for (j = 0; j < iNumRuns; j++)

    } //end of try block
    catch ( modelErr & err )
    {
      GoToNoDataState();
      ExternalErrorHandler( err, m_eSimState, true );
      goto NextBatchGroup;
    }
    catch ( ... )
    {
      modelErr stcErr;
      stcErr.iErrorCode = UNKNOWN;
      strcpy( stcErr.cFunction, "clSimManager::RunBatch" );
      GoToNoDataState();
      ExternalErrorHandler( stcErr, m_eSimState, true );
      goto NextBatchGroup;
    }
    //Label for allowing code in the loop to jump to the next batch group
  NextBatchGroup:;
  } //end of for (i = 0; i < iNumBatchGroups; i++)

  //Delete the par file array
  delete[] p_parFileList;
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// CreateGrid()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
clGridBase * clSimManager::CreateGrid( char * cGridName, short int iNumIntVals, short int iNumFloatVals,
     short int iNumCharVals, short int iNumBoolVals, float fXCellLength, float fYCellLength )
     {
       try
       {
         //Make sure the grid manager has been created - if not, return NULL
         if ( !mp_oGridManager ) return NULL;
         return mp_oGridManager->CreateGrid( cGridName, iNumIntVals, iNumFloatVals, iNumCharVals, iNumBoolVals,
              fXCellLength, fYCellLength );
       }
       catch ( modelErr & err )
       {
         GoToNoDataState();
         throw( err );
       }
       catch ( modelMsg & msg )
       { //non-fatal error
         throw( msg );
       }
       catch ( ... )
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy( stcErr.cFunction, "clSimManager::CreateGrid" );
         GoToNoDataState();
         throw( stcErr );
       }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// Object Number Getters
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
int clSimManager::GetNumberOfBehaviors()
{
  try
  {
    if ( NULL == mp_oBehaviorManager ) return 0;
    else
      return mp_oBehaviorManager->GetNumberOfObjects();
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::GetNumberOfBehaviors" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------
int clSimManager::GetNumberOfGrids()
{
  try
  {
    if ( NULL == mp_oGridManager ) return 0;
    else
      return mp_oGridManager->GetNumberOfObjects();
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::GetNumberOfGrids" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------
int clSimManager::GetNumberOfPopulations()
{
  try
  {
    if ( NULL == mp_oPopulationManager ) return 0;
    else
      return mp_oPopulationManager->GetNumberOfObjects();
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimManager::GetNumberOfPopulations" );
    throw( stcErr );
  }
}
//----------------------------------------------------------------------------





